#if defined(UNICODE) && !defined(_UNICODE)
    #define _UNICODE
#elif defined(_UNICODE) && !defined(UNICODE)
    #define UNICODE
#endif

#include <stdio.h>
#include <tchar.h>
#include <windows.h>
#include <vector>
using namespace std;

/*  Declare Windows procedure  */
LRESULT CALLBACK WindowProcedure (HWND, UINT, WPARAM, LPARAM);

/*  Make the class name into a global variable  */
TCHAR szClassName[ ] = _T("CodeBlocksWindowsApp");

HWND hwndNextViewer;
vector<WCHAR> messageText;

int WINAPI WinMain (HINSTANCE hThisInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR lpszArgument,
                     int nCmdShow)
{
    HWND hwnd;               /* This is the handle for our window */
    MSG messages;            /* Here messages to the application are saved */
    WNDCLASSEX wincl;        /* Data structure for the windowclass */

    /* The Window structure */
    wincl.hInstance = hThisInstance;
    wincl.lpszClassName = szClassName;
    wincl.lpfnWndProc = WindowProcedure;      /* This function is called by windows */
    wincl.style = CS_DBLCLKS;                 /* Catch double-clicks */
    wincl.cbSize = sizeof (WNDCLASSEX);

    /* Use default icon and mouse-pointer */
    wincl.hIcon = LoadIcon (NULL, IDI_APPLICATION);
    wincl.hIconSm = LoadIcon (NULL, IDI_APPLICATION);
    wincl.hCursor = LoadCursor (NULL, IDC_ARROW);
    wincl.lpszMenuName = NULL;                 /* No menu */
    wincl.cbClsExtra = 0;                      /* No extra bytes after the window class */
    wincl.cbWndExtra = 0;                      /* structure or the window instance */
    /* Use Windows's default colour as the background of the window */
    wincl.hbrBackground = (HBRUSH) COLOR_BACKGROUND;

    /* Register the window class, and if it fails quit the program */
    if (!RegisterClassEx (&wincl))
        return 0;

    /* The class is registered, let's create the program*/
    hwnd = CreateWindowEx (
           0,                   /* Extended possibilites for variation */
           szClassName,         /* Classname */
           _T("Clipboard UTF-8 Repairer"),       /* Title Text */
           WS_OVERLAPPEDWINDOW, /* default window */
           CW_USEDEFAULT,       /* Windows decides the position */
           CW_USEDEFAULT,       /* where the window ends up on the screen */
           480,                 /* The programs width */
           64,                  /* and height in pixels */
           HWND_DESKTOP,        /* The window is a child-window to desktop */
           NULL,                /* No menu */
           hThisInstance,       /* Program Instance handler */
           NULL                 /* No Window Creation data */
           );

    /* Make the window visible on the screen */
    ShowWindow (hwnd, nCmdShow);
    hwndNextViewer = SetClipboardViewer (hwnd);

    {
        DWORD pid;
        GetWindowThreadProcessId(hwndNextViewer, &pid);
        char wTextBuffer[1024];
        GetWindowText(hwndNextViewer, wTextBuffer, 1024);
        //fprintf(stderr, "Next clipboard handler is %s (pid: %lu)\n", wTextBuffer, pid);

    }

    /* Run the message loop. It will run until GetMessage() returns 0 */
    while (GetMessage (&messages, NULL, 0, 0))
    {
        /* Translate virtual-key messages into character messages */
        TranslateMessage(&messages);
        /* Send message to WindowProcedure */
        DispatchMessage(&messages);
    }

    /* The program return-value is 0 - The value that PostQuitMessage() gave */
    return messages.wParam;
}


/*  This function is called by the Windows function DispatchMessage()  */

LRESULT CALLBACK WindowProcedure (HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    PAINTSTRUCT ps;
    HDC hdc;

    switch (message)                  /* handle the messages */
    {
        case WM_DESTROY:
            ChangeClipboardChain(hwnd, hwndNextViewer);
            PostQuitMessage (0);       /* send a WM_QUIT to the message queue */
            break;
        case WM_PAINT:
            hdc = BeginPaint(hwnd, &ps);
            FillRect(hdc, &ps.rcPaint, (HBRUSH) (COLOR_WINDOW+1));
            //fprintf(stderr, "Displaying text: ");
            //for (size_t i = 0; i < messageText.size(); i++) {
            //    fprintf(stderr, "%x ", messageText[i]);
            //}
            //fprintf(stderr, "\n");
            //fflush(stderr);
            TextOutW(hdc, 0, 0, messageText.data(), messageText.size() - 1);
            EndPaint(hwnd, &ps);
            break;
        case WM_DRAWCLIPBOARD:
            //fprintf(stderr, "Got a clipboard message\n");
            if (OpenClipboard(NULL))
            {
                HANDLE hClipboardData = GetClipboardData(CF_UNICODETEXT);
                if (hClipboardData)
                {
                    WCHAR *pchData = (WCHAR*) GlobalLock(hClipboardData);
                    if (pchData)
                    {
                        //fprintf(stderr, "Got unicode data\n");
                        vector<WCHAR> converted;
                        WCHAR *ptr = pchData;
                        WCHAR currentUnicodeChar = 0;
                        int unicodeChars = 0;
                        while (*ptr)
                        {
                            //fprintf(stderr, "%x ", *ptr);
                            WCHAR byte1 = *ptr & 0xFF;
                            WCHAR byte2 = (*ptr >> 8) & 0xFF;
                            if (*ptr > 0xFF) {
                                // First two bytes of UTF-8 triplet
                                if (currentUnicodeChar == 0 && byte1 >= 0x60 && byte1 < 0xF0 && byte2 >= 0x80 && byte2 < 0xC0) {
                                    currentUnicodeChar = (byte1 & 0x0F) << 12;
                                    currentUnicodeChar |= (byte2 & 0x3F) << 6;
                                } else {
                                    currentUnicodeChar = 0;
                                }
                            } else {
                                // Last byte of UTF-8 triplet
                                if (currentUnicodeChar > 0 && byte1 >= 0x80 && byte1 < 0xC0) {
                                    currentUnicodeChar |= byte1 & 0x3F;
                                    converted.push_back(currentUnicodeChar);
                                    unicodeChars++;
                                } else {
                                    converted.push_back(byte1);
                                }
                                currentUnicodeChar = 0;
                            }
                            ptr++;
                        }
                        if (unicodeChars > 0) {
                            //fprintf(stderr, "\nConverted to:\n");

                            //for (size_t i = 0; i < converted.size(); i++) {
                            //    fprintf(stderr, "%x ", converted[i]);
                            //}
                            //fprintf(stderr, "\n");
                            GlobalUnlock(hClipboardData);

                            // Okay, let's copy in our new string
                            converted.push_back(0); // terminator
                            messageText = converted;
                            size_t memSize = (converted.size()) * sizeof(WCHAR);
                            HANDLE newclipboard_hData = GlobalAlloc(GMEM_MOVEABLE | GMEM_DDESHARE, memSize);
                            memcpy(GlobalLock(newclipboard_hData), converted.data(), memSize);
                            GlobalUnlock(newclipboard_hData);
                            EmptyClipboard();
                            SetClipboardData(CF_UNICODETEXT,newclipboard_hData);
                            RedrawWindow(hwnd, NULL, NULL, RDW_INVALIDATE);
                        }
                    }
                }
                CloseClipboard();
            }


            SendMessage(hwndNextViewer, message, wParam, lParam);
            //fflush(stderr);
            break;
        default:                      /* for messages that we don't deal with */
            return DefWindowProc (hwnd, message, wParam, lParam);
    }

    return 0;
}
